
/*
 * get_data.c - get data (image header and body)
 */

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "sentry.h"

#include <errno.h>

char *buff = NULL;

/*
 * get_data - whenever data packets are arrived from server,
 *            this procedure will be called.
 */
int
get_data(int datafd)
{
  struct sockaddr_in from;
  int fromlen;
  char packet[8192];
  int nread;
  sentry_image_header *hdr;
  int chk;
  struct timeval image_tstamp;
  int image_format, image_size, image_width, image_height;

  /*
   * receive packets 
   */
  while (1) {
    fromlen = sizeof(from);
    /* bzero(packet, sizeof(packet));*/

    if (debug) {
      time_t now;
      time(&now);
      fprintf(stderr, "waiting for packets... %s\n", ctime(&now));
    }
    nread = recvfrom(datafd, packet, sizeof(packet), 0, (struct sockaddr *)&from, &fromlen);
    
    if (nread < 0) {
      if (errno == EINTR) 
	continue;

      perror("get_data: fail recvfrom()");
      return 1;
    }

    if (debug)
      fprintf(stderr, "get_data: receive %d bytes from %s:%d\n",
	      nread, inet_ntoa(from.sin_addr), htons(from.sin_port));
    
    /*
     * header check
     */
    if ((nread -= SENTRY_IMAGE_HEADER_LEN) < 0) {
      fprintf(stderr, "get_data: header length is wrong.\n");
      return 1;
    }

    hdr = (sentry_image_header *)packet;
    
    if (hdr->hd_ver != SENTRY_VERSION) {
      fprintf(stderr, "get_data: header version is wrong.\n");
      return 1;
    }

#if 1
    if (debug) {
      static int c=0;
      char filename[256];

      sprintf(filename, "./save-file/sample%d.jpg", c);
      if (++c >= 7) c -= 7;

      dump_to_file(filename, &packet[SENTRY_IMAGE_HEADER_LEN], nread);
    }
#endif

    chk = udp_reassemble(hdr, &packet[SENTRY_IMAGE_HEADER_LEN], nread, &buff);

#if 1
    if (debug) {
      char filename[] = "./save-file/debug-sample.jpg";

      fprintf(stderr, "# chk = %d\n", chk);
      dump_to_file(filename, buff, chk);
    }
#endif

    if (chk > 0) {
      fprintf(stdout, ".\n");
      break;
    }

    if (chk == 0) {
      /* need more udp packets, to construct image frame */
      fprintf(stdout, ".");
      continue;
    } 
    else /* chk < 0 */
      return 1;

  } /* end-of-while */
    
  /* 
   * check timestamp ?
   */
  

  /* no-op */
  image_tstamp.tv_sec = ntohl(hdr->hd_tstamp.tv_sec);
  image_tstamp.tv_usec = ntohl(hdr->hd_tstamp.tv_usec);
  image_format = ntohl(hdr->format);
  image_size = ntohl(hdr->size);
  image_width = ntohl(hdr->width);
  image_height = ntohl(hdr->height);
  
  /*
   * save image to shared memory as it is, 
   *   reagardless of image format.
   */
  save_thru(image_tstamp, image_size, image_format, &buff);
  
  return 0;
}
