/*
 * Sentry Video System
 *   is video server with QoS control from clients' requests.
 *
 *      original by akira-y@is.aist-nara.ac.jp, 1994,1995
 *      modified by k-chinen@is.aist-nara.ac.jp, 1996
 *      modified by satosi-f@is.aist-nara.ac.jp, 1997,2000
 *  
 *  $Id: $
 */

/*
 * get_request.c - get request
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <signal.h>

#include "sentry.h"
#include "sentryd.h"

#include <errno.h>

extern int errno;
extern int debug;
extern int verb;

extern int udpsock;
extern int connect_host();

_req_val req_val;

char recvbuf[8192];
struct itimerval sendtimer;

void change_video_qos();
void change_network_qos();
void set_rate();
void set_qos();

extern  int     maxwritefds;
extern  fd_set  writemask;

int 
get_udp_request(struct sockaddr_in *from, int *fromlen)
{
  int nr;
  sentry_request_hdr *sentry_req_hdr;
  u_char type;
  int sockfd;
  int chk;
  extern char *vserver;

  while (1) {
    nr = recvfrom(udpsock, recvbuf, sizeof(recvbuf), 0, (struct sockaddr *)from, fromlen);
    if (nr > 0) break;
    if (errno == EINTR)
      continue;
    fprintf(stderr, "get_udp_request: recvfrom: %s.\n", sys_errlist[errno]);
    return -1;
  }
  
  /* header check */
  sentry_req_hdr = (sentry_request_hdr *)recvbuf;
  if (sentry_req_hdr->hd_ver != SENTRY_VERSION) {
    fprintf(stderr, "get_udp_request: invalid protocol# %d.\n", sentry_req_hdr->hd_ver);
    return -1;
  }
    
  type = sentry_req_hdr->type;

  /* switch(type) {*/
  while (type) {
    if (debug)
      fprintf(stderr, "get_udp_request: type# %x\n", type);
    
    if (type == OPEN_UDP_UNICAST) {
      /* case OPEN_UDP_UNICAST: */
      if (verb)
	fprintf(stdout, "OPEN_UDP_UNICAST request from %s:%d, dataport# %d\n",
		inet_ntoa(from->sin_addr),
		ntohs(from->sin_port), ntohs(sentry_req_hdr->port));
      
      chk = find_client(from);
      if (debug) 
	fprintf(stderr, "get_udp_request: chk(find_client) = %d\n", chk);
      if (chk != -1) break;
      
      
      
      if ((sockfd = connect_host((from->sin_addr), (int)sentry_req_hdr->port, SOCK_DGRAM)) < 0) {
	fprintf(stderr, "sentryd: cannot create data port.\n");
	return -1;
      }
      
      if (debug) 
	fprintf(stderr, "get_udp_request: sockfd = %d\n", sockfd);
      
      if (add_client(type, from, sockfd) < 0) {
#if 0
      chk = add_client(type, from, sockfd);
      if (chk == -2) return 0; /* already registerd in client_list */
      if (chk < -1) {
#endif
	fprintf(stderr, "sentryd: cannot add a new client.\n");
	close(sockfd);
	return -1;
      } 

      FD_SET(sockfd, &writemask);
      if (maxwritefds < sockfd) maxwritefds = sockfd;

      if (!check_video(vserver)) init_video(vserver);

      set_qos(QOS_SIZE, ntohl(sentry_req_hdr->size));
      set_qos(QOS_ZOOM, ntohl(sentry_req_hdr->zoom));
      set_qos(QOS_COLOR, ntohl(sentry_req_hdr->color));;
      set_qos(QOS_FORMAT, ntohl(sentry_req_hdr->format));
      /* change_video_qos(); */

      set_qos(QOS_RATE, ntohl(sentry_req_hdr->rate));
      /* change_network_qos(); */

      break;

    } else if (type == OPEN_UDP_MULTICAST) {
      /* case OPEN_UDP_MULTICAST: */
      if (verb) 
	fprintf(stdout, "OPEN_UDP_MULTICAST request from %s:%d\n",
		inet_ntoa(from->sin_addr),
		/*inet_ntoa((struct in_addr *)ntohl(from->sin_addr.s_addr)),*/
		ntohs(from->sin_port));

      chk = add_client(type, from, 0);
      if (chk == -2) return 0; /* already registerd in client_list */
      if (chk < 0) {
	fprintf(stderr, "sentryd: cannot add a new client.\n");
	return -1;
      }

      if (!check_video(vserver)) init_video(vserver);
      
      set_qos(QOS_SIZE, ntohl(sentry_req_hdr->size));
      set_qos(QOS_ZOOM, ntohl(sentry_req_hdr->zoom));
      set_qos(QOS_COLOR, ntohl(sentry_req_hdr->color));;
      set_qos(QOS_FORMAT, ntohl(sentry_req_hdr->format));
      /* change_video_qos(); */
      
      set_qos(QOS_RATE, ntohl(sentry_req_hdr->rate));
      /* change_network_qos(); */

      break;

    } else if (type == OPEN_TCP) {
      /* case OPEN_TCP: */
      fprintf(stderr, "sentryd: OPEN_TCP request at udp port is rejected.\n");
      return -1;
    }
      
    /* case CLOSE: */
    if (type == CLOSE) {
      clear_client(find_client(from));
      return 0;
    }

    /* case I_AM_ALIVE: */
    if (type == I_AM_ALIVE) {
      /* no-op: keep present */
      return 0;
    }
    

    /*  case QOS_REQUEST_SIZE: */
    if (type & QOS_REQUEST_SIZE) {
      set_qos(QOS_SIZE, ntohl(sentry_req_hdr->size));
      type &= ~QOS_REQUEST_SIZE;
    }

    /* case QOS_REQUEST_ZOOM: */
    if (type & QOS_REQUEST_ZOOM) {
      set_qos(QOS_ZOOM, ntohl(sentry_req_hdr->zoom));
      type &= ~QOS_REQUEST_ZOOM;
    }

    /*  case QOS_REQUEST_COLOR: */
    if (type & QOS_REQUEST_COLOR) {
      set_qos(QOS_COLOR, ntohl(sentry_req_hdr->color));
      type &= ~QOS_REQUEST_COLOR;
    }
    
    /*  case QOS_REQUEST_FORMAT: */
    if (type & QOS_REQUEST_FORMAT) {
      set_qos(QOS_FORMAT, ntohl(sentry_req_hdr->format));
      type &= ~QOS_REQUEST_FORMAT;
    }
    
    /* case QOS_REQUEST_RATE: */
    if (type & QOS_REQUEST_RATE) {
      set_qos(QOS_RATE, ntohl(sentry_req_hdr->rate));
      type &= ~QOS_REQUEST_RATE;
    }
    /* default: */
    else {
      fprintf(stderr, "sentryd: unknown request type. ignored.\n");
      break;
    }
  } /* end-of "type check" */
  
  change_video_qos();
  change_network_qos();

  return 0;
}
  
/*
 * QoS request control 
 */
static u_char nqflag = QOS_REQUEST_RATE;

void
change_network_qos() 
{
#if 0
  if (debug)
    fprintf(stderr, "change_network_qos: req_val.change_flagi# %x, nqflag# %x\n", 
		req_val.change_flag, nqflag);
#endif

#if 0
  if (req_val.change_flag & nqflag)
#endif
    set_rate();
}

void
set_rate()
{
  int rate;
  int interval;
  int sec, usec;
  struct timeval tv;
  extern void send_image();

  signal(SIGALRM, SIG_IGN);

  rate = GET_QOS(QOS_RATE);
#if 1
  rate = 200;
#endif
  interval = 100 * 1000000 / rate;
  sec = interval / 1000000;
  usec = interval % 1000000;

  if (debug)
    fprintf(stderr, "set_rate: rate# %d, interval# %d,%06d\n", 
	    rate, sec, usec);

  sendtimer.it_interval.tv_sec = sec;
  sendtimer.it_interval.tv_usec = usec;
  sendtimer.it_value.tv_sec = sec;
  sendtimer.it_value.tv_usec = usec;

  signal(SIGALRM, send_image);
  setitimer(ITIMER_REAL, &sendtimer, NULL);

  CLR_CHANGE_FLAG(QOS_RATE);
}

void
set_qos(int type, int val)
{
  int oldval;

  if (debug)
    fprintf(stderr, "set_qos: type[%d] = %d\n", type, val);

  if (type < 0 || type >= QOS_PARAMS) 
    return;

  if (req_val.attr.val[type] != val) {
    oldval = val;
    req_val.attr.val[type] = val;
    req_val.change_flag |= (0x01 << type);

    if (verb)
      fprintf(stderr, "change QoS params# %d, %d -> %d\n", 
	      type, oldval, req_val.attr.val[type]);
      /*fprintf(stderr, "req_val.change_flag = %d\n");*/
  }
}

void 
init_req_val(void)
{
  int type;
  extern int require_encode_type;

  for(type=0; type<QOS_PARAMS; type++) {
    req_val.attr.val[type] = 0;
    if (verb)
      fprintf(stderr, "init_req_val: type# %d initialized.\n", type);
  }

  req_val.attr.val[QOS_FORMAT] = require_encode_type;
  req_val.change_flag = 0;

}

