/*
 * Sentry Video System for PC with video4linux
 *   is controlled QoS according to clients' requests.
 *
 *        original by akira-y@is.aist-nara.ac.jp, 1994,1995
 *        modified by k-chinen@is.aist-nara.ac.jp, 1996
 *        modified by satosi-f@is.aist-nara.ac.jp, 1997,2000
 *        modified by atsush-m@is.aist-nara.ac.jp 2002
 *
 *   $Id: $
 */

/*
 * video.c - video (chunk of images) handling procedures
 */
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <linux/types.h>
#include <linux/videodev.h>

#include "sentry.h"
#include "sentryd.h"

#define VDEV_NAME "/dev/video0"
#define PASS fprintf(stderr,"PASS %s:%d\n", __FILE__, __LINE__)

extern int debug;
extern int verb;
extern int errno;
extern int xsize, ysize;
extern _req_val req_val;

int n, vfd=-1;
char *cur=NULL, *prev=NULL, *mapbase=NULL;
char *vdev_name = VDEV_NAME;

struct video_capability vd;
struct video_channel vc[10];
struct video_picture vp;
struct video_mbuf vm;
struct video_mmap vmm;
struct flock lock;

int
check_video(char *vserver)
{
  /* check if it can open device */
  if(vfd == -1){
    if((vfd = open(vdev_name, O_RDWR)) < 2){
      perror("open video");
      return 1;
    }else{
      close(vfd);
      vfd = -1;
      return 0;
    }
  }

  /* check if device is running */
  fcntl(vfd, F_GETLK, lock);
  if(lock.l_type == F_UNLCK)
    return 0;
  else
    return 1;
}

void 
init_video(char *vserver)
{
  extern void setup_video();

  /* connect to the daemon */
  if((vfd = open(vdev_name, O_RDWR)) < 2){
    perror("open video");
    exit(21);
  }

  /* device lock */
  lock.l_type = F_WRLCK;
  lock.l_start = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len = 0;
  fcntl(vfd, F_SETLK, &lock);

  /* get divice parameter */
  if(ioctl(vfd, VIDIOCGCAP, &vd) < 0){
    perror("ioctl(VIDOCGCAP)");
    exit(22);
  } 

  /* check device parameter */
  fprintf(stderr, "vd.name: \"%s\"\n", vd.name);
  fprintf(stderr, "vd.type=0x%08x", vd.type);
  if(vd.type&VID_TYPE_CAPTURE) fprintf(stderr, " CAPTURE");
  if(vd.type&VID_TYPE_TUNER) fprintf(stderr, " TUNER");
  if(vd.type&VID_TYPE_TELETEXT) fprintf(stderr, " TELETEXT");
  if(vd.type&VID_TYPE_OVERLAY) fprintf(stderr, " OVERLAY");
  if(vd.type&VID_TYPE_CHROMAKEY) fprintf(stderr, " CHROMAKEY");
  if(vd.type&VID_TYPE_CLIPPING) fprintf(stderr, " CLIPPING");
  if(vd.type&VID_TYPE_FRAMERAM) fprintf(stderr, " FRAMERAM");
  if(vd.type&VID_TYPE_SCALES) fprintf(stderr, " SCALES");
  if(vd.type&VID_TYPE_MONOCHROME) fprintf(stderr, " MONOCHROME");
  if(vd.type&VID_TYPE_SUBCAPTURE) fprintf(stderr, " SUBCAPTURE");
  fprintf(stderr, "\n");
  fprintf(stderr, "vd.channels=%d\n", vd.channels);
  fprintf(stderr, "vd.audios=%d\n", vd.audios);
  fprintf(stderr, "vd.maxwidth=%d\n", vd.maxwidth);
  fprintf(stderr, "vd.maxheight=%d\n", vd.maxheight);
  fprintf(stderr, "vd.minwidth=%d\n", vd.minwidth);
  fprintf(stderr, "vd.minheight=%d\n", vd.minheight);
  fprintf(stderr, "\n");

  /* get channel parameters */
  for(n=0; n<vd.channels; n++){
    vc[n].channel = n;
    if(ioctl(vfd, VIDIOCGCHAN, &vc[n]) < 0){
      perror("ioctl(VIDIOCGCHAN)");
      exit(23);
    }

    fprintf(stderr, "vc[%d].channel=%d\n", n, vc[n].channel);
    fprintf(stderr, "vc[%d].name=\"%s\"\n", n, vc[n].name);
    fprintf(stderr, "vc[%d].tuners=%d\n", n, vc[n].tuners);
    fprintf(stderr, "vc[%d].flags=0x%08x", n, vc[n].flags);
    if(vc[n].flags&VIDEO_VC_TUNER) fprintf(stderr, " TUNER");
    if(vc[n].flags&VIDEO_VC_AUDIO) fprintf(stderr, " AUDIO");
    fprintf(stderr, "\n");
    fprintf(stderr, "vc[%d].type=0x%08x", n, vc[n].type);
    if(vc[n].type&VIDEO_TYPE_TV) fprintf(stderr, " TV");
    if(vc[n].type&VIDEO_TYPE_CAMERA) fprintf(stderr, " CAMERA");
    fprintf(stderr, "\n");
    fprintf(stderr, "vc[%d].norm=%d\n", n, vc[n].norm);
    fprintf(stderr, "\n");
  }

  /* choice "Composite1" */
  vc[1].norm = 1;    /* 0:Television 1:Composite1 2:S-Video */  
  if(ioctl(vfd, VIDIOCSCHAN, &vc[1]) < 0){
    perror("ioctl(VIDIOCSCHAN)");
    exit(24);
  }

  (void)setup_video();

}

void
setup_video(void)
{
  extern void set_color(), set_zoom(), set_size();

  /* get mmap information */
  if(ioctl(vfd, VIDIOCGMBUF, &vm) < 0){
    perror("ioctl(VIDIOCGMBUF)");
    exit(25);
  }

  if((mapbase = mmap(0, vm.size, PROT_READ|PROT_WRITE, MAP_SHARED, vfd, 0)) == (char *)-1){
    perror("mmap");
    exit(26);
  }
  cur = mapbase + vm.offsets[0];

  if (CHK_CHANGE_FLAG(QOS_SIZE)) set_size();
  if (CHK_CHANGE_FLAG(QOS_ZOOM)) set_zoom();
  if (CHK_CHANGE_FLAG(QOS_COLOR)) set_color();
}

char *
capture_frame()
{
  extern void convert_color();

  if (debug)
    fprintf(stderr, "capture_frame:\n");

  /* capture start */
  vmm.frame = 0;
  vmm.width = xsize;
  vmm.height = ysize;
  vmm.format = vp.palette;
  if(ioctl(vfd, VIDIOCMCAPTURE, &vmm) < 0){
    perror("ioctl(VIDIOCMCAPTURE)");
    exit(27);
  }

  /* waiting for capture finished */
  if(ioctl(vfd, VIDIOCSYNC, &vmm.frame) < 0){
    perror("ioctl(VIDIOCSYNC)");
    exit(28);
  }

  convert_color();

#if 1
    {
      static int sn=0;
      char fname[BUFSIZ];
      char head[BUFSIZ];
      int fd;
      int nw;
      char *p;
      int c;

      sprintf(fname,"/var/tmp/x%d", (sn++)%7);
      fd = open(fname, O_WRONLY|O_CREAT, 0755);
      if(fd<0) {
        fprintf(stderr,"fail open (%d)\n", errno);
      }
  #if 0
      /* PGM */
      sprintf(head,"P5 %d %d %d\n", xsize, ysize, 255);
      nw = write(fd, head, strlen(head));
      nw = write(fd, cur, xsize*ysize*3);
  #endif
  #if 1
      /* PPM */
      sprintf(head,"P6 %d %d %d\n", xsize, ysize, 255);
      nw = write(fd, head, strlen(head));
      nw = write(fd, cur, xsize*ysize*3);
  #endif
  #if 1
      /* PPM */
      sprintf(head,"P6 %d %d %d\n", xsize, ysize, 255);
      nw = write(fd, head, strlen(head));
      c = 0;
      p = cur;
      while(c<xsize*ysize) {
        nw = write(fd, p, 3);
        p += 4;
        c++;
      }
  #endif
      close(fd);
    }
#endif

  if (cur != prev) {
    if (debug) {
      fprintf(stderr, "capture_frame: prev %d -> cur %d\n", prev, cur);
    }
    /*if (prev) free(prev);*/
    prev = cur;
  }
  return cur;
}

int
release_frame()
{
  extern char *cur;
      
  /*
   * Finished with frame, unlocked the buffer
   */
  if (debug)
    fprintf(stderr, "release_frame:\n");

  return 0;
}


/*
 * set video capture parameters 
 */
void
set_size(void)
{
  int req;

  req = GET_QOS(QOS_SIZE);
#if 1
  req = SIZE_160x120;
#endif

  /* Set the frame size
   *   640x480, 320x240, 212x160, 160x120: normal vision (w:h =  4:3)
   *   640x360, 320x180, 212x120, 160x90: wide vision (w:h = 16:9 )
   */
#if 0
  switch (GET_QOS(QOS_SIZE)) {
#endif
  switch (req) {
  case SIZE_640x480:
    xsize = 640;
    ysize = 480;
    break;
  case SIZE_320x240:
    xsize = 320;
    ysize = 240;
    break;
  case SIZE_212x160:
  default:
    xsize = 212;
    ysize = 160;
    break;
  case SIZE_160x120:
    xsize = 160;
    ysize = 120;
    break;
    
  case SIZE_640x360:
    xsize = 640;
    ysize = 360;
    break;
  case SIZE_320x180:
    xsize = 320;
    ysize = 180;
    break;
  case SIZE_212x120:
    xsize = 212;
    ysize = 120;
    break;
    
  case SIZE_160x90:
    xsize = 160;
    ysize = 90;
    break;
  }

  CLR_CHANGE_FLAG(QOS_SIZE);
}

void
set_zoom(void)
{
  CLR_CHANGE_FLAG(QOS_ZOOM);
}

void 
set_color(void)
{
  PASS;
 
  /* set the packing the RGB. Can't use YUV  */
  vp.brightness = 32767;
  vp.hue = 32767;
  vp.colour = 32767;
  vp.contrast = 32767;
  vp.whiteness = 32767;
  switch (GET_QOS(QOS_COLOR)) {
  case FULL_COLOR:
     vp.depth = 24;
     vp.palette = VIDEO_PALETTE_RGB24;
     break;
  case MONOCHROME:
  default:
     vp.palette = VIDEO_PALETTE_GREY;
    break;
  }
  if(ioctl(vfd, VIDIOCSPICT, &vp)){
    perror("ioctl(VIDIOCSPICT)");
    exit(28);
  }

  CLR_CHANGE_FLAG(QOS_COLOR);
}

int 
close_video()
{
    if (verb)
      fprintf(stderr, "close_video:\n");

    munmap((void *)cur, vm.size);

    return 0;
}

/*
 * change_video_qos: change capturing parameter.
 *   this function is realized by re-setup_video(),
 *   which may cause the memory leak. (future work:-p)
 */
static u_char vqflag = 
(QOS_REQUEST_SIZE | QOS_REQUEST_ZOOM | QOS_REQUEST_COLOR | QOS_REQUEST_FORMAT);

void
change_video_qos()
{
#if 0
  if (debug)
    fprintf(stderr, "change_video_qos: change_flag# %x, vqflag# %x\n",
		req_val.change_flag, vqflag);
#endif

  if (req_val.change_flag & vqflag) {
    close_video();
    setup_video();
  }
}

void
convert_color()
{
  char *ptr1, *ptr2, tmp;

  /* covert RGB lines. BGRx -> RGB */
  for(n=0, ptr1=cur, ptr2=cur; n<xsize*ysize; n++, ptr1+=4, ptr2+=3){
    tmp=ptr1[0]; ptr2[0]=ptr1[2]; ptr2[1]=ptr1[1]; ptr2[2]=tmp;
  }
}
